/*
Copyright (c) 2013, BeOne Stuttgart GmbH
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
- Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer
 in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.beone.sample;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.jface.layout.GridDataFactory;
import org.eclipse.nebula.widgets.nattable.NatTable;
import org.eclipse.nebula.widgets.nattable.config.AbstractRegistryConfiguration;
import org.eclipse.nebula.widgets.nattable.config.CellConfigAttributes;
import org.eclipse.nebula.widgets.nattable.config.DefaultNatTableStyleConfiguration;
import org.eclipse.nebula.widgets.nattable.config.IConfigRegistry;
import org.eclipse.nebula.widgets.nattable.config.IEditableRule;
import org.eclipse.nebula.widgets.nattable.data.IColumnPropertyAccessor;
import org.eclipse.nebula.widgets.nattable.data.IDataProvider;
import org.eclipse.nebula.widgets.nattable.data.ListDataProvider;
import org.eclipse.nebula.widgets.nattable.data.convert.DefaultBooleanDisplayConverter;
import org.eclipse.nebula.widgets.nattable.data.convert.DefaultIntegerDisplayConverter;
import org.eclipse.nebula.widgets.nattable.edit.EditConfigAttributes;
import org.eclipse.nebula.widgets.nattable.edit.editor.CheckBoxCellEditor;
import org.eclipse.nebula.widgets.nattable.edit.editor.ComboBoxCellEditor;
import org.eclipse.nebula.widgets.nattable.edit.editor.TextCellEditor;
import org.eclipse.nebula.widgets.nattable.grid.GridRegion;
import org.eclipse.nebula.widgets.nattable.grid.data.DefaultColumnHeaderDataProvider;
import org.eclipse.nebula.widgets.nattable.grid.layer.DefaultGridLayer;
import org.eclipse.nebula.widgets.nattable.layer.DataLayer;
import org.eclipse.nebula.widgets.nattable.layer.ILayer;
import org.eclipse.nebula.widgets.nattable.layer.LabelStack;
import org.eclipse.nebula.widgets.nattable.layer.cell.ColumnOverrideLabelAccumulator;
import org.eclipse.nebula.widgets.nattable.painter.cell.CheckBoxPainter;
import org.eclipse.nebula.widgets.nattable.selection.SelectionLayer;
import org.eclipse.nebula.widgets.nattable.selection.action.AbstractMouseSelectionAction;
import org.eclipse.nebula.widgets.nattable.selection.config.DefaultSelectionBindings;
import org.eclipse.nebula.widgets.nattable.selection.config.DefaultSelectionLayerConfiguration;
import org.eclipse.nebula.widgets.nattable.style.CellStyleAttributes;
import org.eclipse.nebula.widgets.nattable.style.DisplayMode;
import org.eclipse.nebula.widgets.nattable.style.Style;
import org.eclipse.nebula.widgets.nattable.tooltip.NatTableContentTooltip;
import org.eclipse.nebula.widgets.nattable.ui.binding.UiBindingRegistry;
import org.eclipse.nebula.widgets.nattable.ui.matcher.MouseEventMatcher;
import org.eclipse.nebula.widgets.nattable.ui.menu.IMenuItemProvider;
import org.eclipse.nebula.widgets.nattable.ui.menu.PopupMenuBuilder;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.ui.part.ViewPart;

import com.beone.sample.table.SampleDataElement;

/**
 * This class implements the main view of the application, providing the UI controls and logic.
 *
 */
public class View extends ViewPart {
	public static final String ID = "SampleView";
	
	private static final String INTEGER_LABEL = "Label_Integer";
	private static final String TEXT_LABEL = "Label_Text";
	private static final String CHECKBOX_LABEL = "Label_Checkbox";
	private static final String COMBOBOX_LABEL = "Label_Combobox";
	
	public static final String LEVEL_BEGINNER = "beginner";
	public static final String LEVEL_INTERMEDIATE = "intermediate";
	public static final String LEVEL_EXPERT = "expert";
	
	public static final String EXPERT_LABEL = "Label_Expert";
	

	private NatTable natTable;
	
	private List<SampleDataElement> elements;

	/**
	 * Create the elements of the view.
	 */
	public void createPartControl(Composite parent) {
		
		parent.setLayout(new GridLayout());
		
		// Initial data for the table
		elements = new ArrayList<>();
		elements.add(new SampleDataElement(1, "Buchwald", "Jan-Paul", LEVEL_INTERMEDIATE, true));
		elements.add(new SampleDataElement(2, "Fauth", "Dirk", LEVEL_EXPERT, true));
		final String[] propertyNames = {
				"Number",
				"Last Name",
				"First Name",
				"Level",
				"Member"
		};
		
		// Add-Button
		Button addButton = new Button(parent, SWT.NONE);
		addButton.setText("Add new table element");
		addButton.addSelectionListener(new SelectionListener() {
			
			@Override
			public void widgetSelected(SelectionEvent e) {
				// add a new entry with default values to the list, refresh table
				elements.add(new SampleDataElement(elements.size()+1, "lastName", "firstName", LEVEL_BEGINNER, false));
				natTable.refresh();
			}
			
			@Override
			public void widgetDefaultSelected(SelectionEvent e) {
				widgetSelected(e);
			}
		});
		
		
		// NatTable
		
		// Create a column property accessor
		IColumnPropertyAccessor<SampleDataElement> columnPropertyAccessor = new IColumnPropertyAccessor<SampleDataElement>() {
			
			@Override
			public String getColumnProperty(int index) {
				return propertyNames[index];
			}
			
			@Override
			public int getColumnIndex(String name) {
				for (int i=0; i < propertyNames.length; i++) {
					if (propertyNames[i].equals(name)) {
						return i;
					}
				}
				return -1;
			}
			
			@Override
			public void setDataValue(SampleDataElement element, int index, Object value) {
				switch (index) {
				case 0:
					element.setNumber((Integer)value);
					break;
				case 1:
					element.setLastName(String.valueOf(value));
					break;
				case 2:
					element.setFirstName(String.valueOf(value));
					break;
				case 3:
					element.setLevel(String.valueOf(value));
					break;
				case 4:
					element.setMember((Boolean)value);
					break;
				}
			}
			
			@Override
			public Object getDataValue(SampleDataElement element, int index) {
				switch (index) {
					case 0:
						return element.getNumber();
					case 1:
						return element.getLastName();
					case 2:
						return element.getFirstName();
					case 3:
						return element.getLevel();
					case 4:
						return element.isMember();
					default:
						return null;
				}
			}
			
			@Override
			public int getColumnCount() {
				return 5;
			}
		};
		
		final IDataProvider bodyDataProvider = new ListDataProvider<SampleDataElement>(elements, columnPropertyAccessor);
		DefaultGridLayer gridLayer = new DefaultGridLayer(
				bodyDataProvider, new DefaultColumnHeaderDataProvider(propertyNames));
		
		// find selection layer and register row header mouse click binding
		ILayer currentLayer = gridLayer.getBodyLayer();
		while (currentLayer != null && !(currentLayer instanceof SelectionLayer)) {
			currentLayer = currentLayer.getUnderlyingLayersByRowPosition(0).iterator().next();
		}
		
		((SelectionLayer)currentLayer).addConfiguration(new DefaultSelectionLayerConfiguration() {
			@Override
			protected void addSelectionUIBindings() {
				addConfiguration(new DefaultSelectionBindings() {
					@Override
					protected void configureRowHeaderMouseClickBindings(
							UiBindingRegistry uiBindingRegistry) {
						// register Action on row header right click
						RowHeaderMouseRightClickAction action = new RowHeaderMouseRightClickAction();
						uiBindingRegistry.registerFirstSingleClickBinding(MouseEventMatcher.rowHeaderRightClick(SWT.NONE), action);
					}
				});
			}
		});
		
		// Labels
		final DataLayer bodyDataLayer = (DataLayer) gridLayer.getBodyDataLayer();
		final ColumnOverrideLabelAccumulator columnLabelAccumulator = new ColumnOverrideLabelAccumulator(bodyDataLayer) {
			@Override
			public void accumulateConfigLabels(LabelStack configLabels,
					int columnPosition, int rowPosition) {
				super.accumulateConfigLabels(configLabels, columnPosition, rowPosition);
				// additionally set label for expert
				if (columnPosition == 3 && "expert".equals(bodyDataProvider.getDataValue(columnPosition, rowPosition))) {
					configLabels.addLabel(EXPERT_LABEL);
				}
			}
		};
		
		bodyDataLayer.setConfigLabelAccumulator(columnLabelAccumulator);
		columnLabelAccumulator.registerColumnOverrides(0, INTEGER_LABEL);
		columnLabelAccumulator.registerColumnOverrides(1, TEXT_LABEL);
		columnLabelAccumulator.registerColumnOverrides(2, TEXT_LABEL);
		columnLabelAccumulator.registerColumnOverrides(3, COMBOBOX_LABEL);
		columnLabelAccumulator.registerColumnOverrides(4, CHECKBOX_LABEL);
		
		// Create NatTable and add configuration objects
		natTable = new NatTable(parent, gridLayer, false);
		natTable.addConfiguration(new DefaultNatTableStyleConfiguration());
		natTable.addConfiguration(new EditorConfiguration());
		natTable.configure();
		
		// Add default content tooltip for column headers and body (displays the cell value in the tooltip)
		new NatTableContentTooltip(natTable, new String[] {GridRegion.COLUMN_HEADER, GridRegion.BODY});
		
		GridDataFactory.fillDefaults().grab(true, true).applyTo(natTable);
	}
	
	/*
	 * Contains registration of editors and special styles based on labels.
	 */
	class EditorConfiguration extends AbstractRegistryConfiguration  {

		@Override
		public void configureRegistry(IConfigRegistry configRegistry) {
			configRegistry.registerConfigAttribute(
					EditConfigAttributes.CELL_EDITABLE_RULE, 
					IEditableRule.ALWAYS_EDITABLE);
			
			registerEditors(configRegistry);
			registerExpertStyle(configRegistry);
		}

		private void registerEditors(IConfigRegistry configRegistry) {
			registerIntegerEditor(configRegistry);
			registerTextEditor(configRegistry);
			registerCheckbox(configRegistry);
			registerCombobox(configRegistry);
		}
		
		/*
		 * Expert Style: Green background if expert label is set
		 */
		private void registerExpertStyle(IConfigRegistry configRegistry) {
			Style expertStyle = new Style();
			expertStyle.setAttributeValue(CellStyleAttributes.BACKGROUND_COLOR, Display.getCurrent().getSystemColor(SWT.COLOR_GREEN));
			configRegistry.registerConfigAttribute(CellConfigAttributes.CELL_STYLE, expertStyle,
					DisplayMode.NORMAL, EXPERT_LABEL);
		}
		
		/*
		 * Text editor: default TextCellEditor for text label.
		 */
		private void registerTextEditor(IConfigRegistry configRegistry) {
			//moves the selection after commit by enter
			configRegistry.registerConfigAttribute(
					EditConfigAttributes.CELL_EDITOR, 
					new TextCellEditor(true, true), 
					DisplayMode.NORMAL, TEXT_LABEL);
		}
		
		/*
		 * Integer Editor: Text Cell Editor and Integer converter for integer label.
		 */
		private void registerIntegerEditor(IConfigRegistry configRegistry) {
			configRegistry.registerConfigAttribute(
					EditConfigAttributes.CELL_EDITOR, 
					new TextCellEditor(false, true), 
					DisplayMode.NORMAL, 
					INTEGER_LABEL);
			
			// Integer converter
			configRegistry.registerConfigAttribute(
					CellConfigAttributes.DISPLAY_CONVERTER, 
					new DefaultIntegerDisplayConverter(), 
					DisplayMode.NORMAL,
					INTEGER_LABEL);
		}
		
		/*
		 * Checkbox Editor: CheckboxCellEditor for checkbox label.
		 */
		private void registerCheckbox(IConfigRegistry configRegistry) {
			configRegistry.registerConfigAttribute(
					EditConfigAttributes.CELL_EDITOR, 
					new CheckBoxCellEditor(), 
					DisplayMode.EDIT, 
					CHECKBOX_LABEL);
			
			// CheckBoxPainter to show the checkbox in the NatTable
			configRegistry.registerConfigAttribute(
					CellConfigAttributes.CELL_PAINTER, 
					new CheckBoxPainter(), 
					DisplayMode.NORMAL, 
					CHECKBOX_LABEL);
			
			// Boolean Converter
			configRegistry.registerConfigAttribute(
					CellConfigAttributes.DISPLAY_CONVERTER, 
					new DefaultBooleanDisplayConverter(), 
					DisplayMode.NORMAL, 
					CHECKBOX_LABEL);
		}
		
		/*
		 * Combo Box for level selection, three hard-coded value options.
		 */
		private void registerCombobox(IConfigRegistry configRegistry) {
			List<String> selectValues = new ArrayList<String>();
			selectValues.add(LEVEL_BEGINNER);
			selectValues.add(LEVEL_INTERMEDIATE);
			selectValues.add(LEVEL_EXPERT);
			ComboBoxCellEditor comboEditor = new ComboBoxCellEditor(selectValues);
			configRegistry.registerConfigAttribute(EditConfigAttributes.CELL_EDITOR,
					comboEditor, DisplayMode.EDIT, COMBOBOX_LABEL);
		}
	}
	
	/*
	 * Right Click Action that produces a context men with an entry to delete
	 * the current row.
	 *
	 */
	class RowHeaderMouseRightClickAction extends AbstractMouseSelectionAction {
		
		@Override
		public void run(final NatTable natTable, final MouseEvent mouseEvent) {
			// create pop-up menu
			PopupMenuBuilder menuBuilder = new PopupMenuBuilder(natTable)
				.withMenuItemProvider(new IMenuItemProvider() {
					
					@Override
					public void addMenuItem(final NatTable table, final Menu popupMenu) {
						// add one menu item
						MenuItem menuItem = new MenuItem(popupMenu, SWT.PUSH);
						menuItem.setText("Delete entry");
						menuItem.setEnabled(true);

						menuItem.addSelectionListener(new SelectionAdapter() {
							@Override
							public void widgetSelected(SelectionEvent event) {
								// current row position is determined by y coordinate of original mouse event
								// -1 due to header row
								elements.remove(table.getRowPositionByY(mouseEvent.y)-1);
								// refresh table and reset context men
								table.refresh();
								popupMenu.setVisible(false);
								natTable.setMenu(null);
							}
						});

					}
				});
			
			Menu rowMenu = menuBuilder.build();
			rowMenu.setVisible(true);
			// Testability: set menu to be able to find it in swtbot test
			natTable.setMenu(rowMenu);
		}
	}

	/**
	 * Passing the focus request to the table.
	 */
	public void setFocus() {
		natTable.setFocus();
	}
}