/*
Copyright (c) 2013, BeOne Stuttgart GmbH
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
- Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer
 in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.beone.sample.test;

import org.eclipse.jface.window.DefaultToolTip;
import org.eclipse.nebula.widgets.nattable.NatTable;
import org.eclipse.nebula.widgets.nattable.command.ILayerCommand;
import org.eclipse.nebula.widgets.nattable.edit.EditConfigAttributes;
import org.eclipse.nebula.widgets.nattable.edit.InlineEditHandler;
import org.eclipse.nebula.widgets.nattable.edit.command.UpdateDataCommand;
import org.eclipse.nebula.widgets.nattable.edit.editor.ComboBoxCellEditor;
import org.eclipse.nebula.widgets.nattable.edit.editor.ICellEditor;
import org.eclipse.nebula.widgets.nattable.grid.GridRegion;
import org.eclipse.nebula.widgets.nattable.layer.CompositeLayer;
import org.eclipse.nebula.widgets.nattable.layer.ILayer;
import org.eclipse.nebula.widgets.nattable.layer.LabelStack;
import org.eclipse.nebula.widgets.nattable.layer.cell.ILayerCell;
import org.eclipse.nebula.widgets.nattable.selection.ISelectionModel;
import org.eclipse.nebula.widgets.nattable.selection.SelectionLayer;
import org.eclipse.nebula.widgets.nattable.selection.SelectionLayer.MoveDirectionEnum;
import org.eclipse.nebula.widgets.nattable.selection.command.SelectCellCommand;
import org.eclipse.nebula.widgets.nattable.selection.command.SelectRowsCommand;
import org.eclipse.nebula.widgets.nattable.style.DisplayMode;
import org.eclipse.nebula.widgets.nattable.widget.EditModeEnum;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swtbot.swt.finder.SWTBot;
import org.eclipse.swtbot.swt.finder.exceptions.WidgetNotFoundException;
import org.eclipse.swtbot.swt.finder.results.ArrayResult;
import org.eclipse.swtbot.swt.finder.results.BoolResult;
import org.eclipse.swtbot.swt.finder.results.IntResult;
import org.eclipse.swtbot.swt.finder.results.Result;
import org.eclipse.swtbot.swt.finder.results.StringResult;
import org.eclipse.swtbot.swt.finder.results.VoidResult;
import org.eclipse.swtbot.swt.finder.waits.DefaultCondition;
import org.eclipse.swtbot.swt.finder.waits.ICondition;
import org.eclipse.swtbot.swt.finder.widgets.AbstractSWTBot;
import org.eclipse.swtbot.swt.finder.widgets.SWTBotMenu;
import org.eclipse.swtbot.swt.finder.widgets.SWTBotShell;

/**
 * SWT Bot Implementation for the Eclipse Nebula NatTable component.
 * Provides convenience methods and conditions that are useful
 * when testing Eclipse based applications containing NatTable table instances.
 */
public class SWTBotNatTable extends AbstractSWTBot<NatTable> {

	public SWTBotNatTable(NatTable w) throws WidgetNotFoundException {
		super(w);
	}
	
	/** Determines and returns the body layer of the table */
	public ILayer getBodyLayer() {
		// Composite Layer (this includes GridLayer): look up body layer by Region Label
		if (widget.getLayer() instanceof CompositeLayer) {
			return findLayerByRegionLabel(GridRegion.BODY);
		}
		// otherwise just return the top layer
		return widget.getLayer();
	}
	
	/**
	 * Returns the row header layer if there is one.
	 */
	public ILayer getRowHeaderLayer() {
		if (widget.getLayer() instanceof CompositeLayer) {
			return findLayerByRegionLabel(GridRegion.ROW_HEADER);
		}
		return null;
	}
	
	/**
	 * Returns the column header layer if there is one.
	 */
	public ILayer getColumnHeaderLayer() {
		if (widget.getLayer() instanceof CompositeLayer) {
			return findLayerByRegionLabel(GridRegion.COLUMN_HEADER);
		}
		return null;
	}
	
	/**
	 * Returns the column group header layer if there is one.
	 */
	public ILayer getColumnGroupHeaderLayer() {
		if (widget.getLayer() instanceof CompositeLayer) {
			return findLayerByRegionLabel(GridRegion.COLUMN_GROUP_HEADER);
		}
		return null;
	}

	/**
	 * Iterates through all layers and returns the first that has the specified regionLabel 
	 * */
	private ILayer findLayerByRegionLabel(String regionLabel) {
		int colcount = widget.getLayer().getColumnCount();
		int rowcount = widget.getLayer().getRowCount();
		for (int col = 0; col < colcount; col++) {
			for (int row = 0; row < rowcount; row++) {
				// Region is stored as a normal label
				if (widget.getLayer().getConfigLabelsByPosition(col, row).hasLabel(regionLabel)) {
					return widget.getLayer().getUnderlyingLayerByPosition(col, row);
				}
			}
		}
		return null;
	}
	
	/** 
	 * Returns the selection model associated to a possible selection layer of the body layer.
	 * Returns null if no such layer is found.
	 */
	public ISelectionModel getBodySelectionModel() {
		ILayer currentLayer = getBodyLayer();
		while (currentLayer != null && !(currentLayer instanceof SelectionLayer)) {
			currentLayer = currentLayer.getUnderlyingLayersByRowPosition(0).iterator().next();
		}
		
		if (currentLayer != null) {
			return ((SelectionLayer)currentLayer).getSelectionModel();
		}
		return null;
	}
	
	/**
	 * Returns the indexes of rows that are currently selected in the body region of the table.
	 */
	public Integer[] getSelectedBodyRows() {
		return syncExec(new ArrayResult<Integer>() {
			@Override
			public Integer[] run() {
				ISelectionModel selectionModel = getBodySelectionModel();
				if (selectionModel != null) {
					int[] selectedRows = selectionModel.getFullySelectedRowPositions(0);
					Integer[] resultArray = new Integer[selectedRows.length];
					for (int i=0; i < selectedRows.length; i++) {
						resultArray[i] = selectedRows[i];
					}
					return resultArray;
				}
				return new Integer[0];
			}
		});
	}

	/**
	 * Triggers the selection of the rows according to the indexes passed in.
	 */
	public void selectRows(final int[] indizes) {
		// synchronous execution of the NatTable select rows command
		syncExec(new VoidResult() {
			@Override
			public void run() {
				widget.doCommand(new SelectRowsCommand(widget, 0, indizes, true, true, 0));				
			}
		});
	}
	
	
	/**
	 * Directly trigger the cell selection on the table instance with the specified key masks.
	 */
	public void selectBodyCell(final int row, final int col, final boolean shift, final boolean ctrl) {
		// synchronous execution of the NatTable select cell command
		syncExec(new VoidResult() {
			@Override
			public void run() {
				ILayer bodyLayer = getBodyLayer();
				bodyLayer.doCommand(new SelectCellCommand(bodyLayer, col, row, shift, ctrl));				
			}
		});
	}

	/**
	 * Executes the given command on the body layer of the table.
	 * The command is executed asynchronously so the UI thread is not blocked.
	 */
	public void executeCommandOnBodyLayer(final ILayerCommand command) {
		asyncExec(new VoidResult() {
			@Override
			public void run() {
				getBodyLayer().doCommand(command);
			}
		});
	}

	
	/**
	 * Returns the current row count of the body region of the table.
	 */
	public int getBodyRowCount() {
		return syncExec(new IntResult() {
			
			@Override
			public Integer run() {
				ILayer bodyLayer = getBodyLayer();
				// return 0 if the body layer is not available
				return bodyLayer!=null?bodyLayer.getPreferredRowCount():0;
			}
		});
	}
	
	/**
	 * Returns the current column count of the body region of the table.
	 */
	public int getBodyColumnCount() {
		return syncExec(new IntResult() {
			
			@Override
			public Integer run() {
				return getBodyLayer().getPreferredColumnCount();
			}
		});
	}

	/**
	 * Returns the value of the given cell in the body region of the table.
	 */
	public Object getBodyValue(final int row, final int col) {
		return syncExec(new Result<Object>() {

			@Override
			public Object run() {
				ILayerCell bodyCell = getBodyLayer().getCellByPosition(col, row);
				if (bodyCell != null) {
					return bodyCell.getDataValue();
				}
				return null;
			}
		});
	}
	
	/**
	 * Returns true if the given cell in the body region has the expected label set,
	 * false otherwise.
	 */
	public boolean hasBodyCellLabel(final int row, final int col, final String label) {
		return syncExec(new BoolResult() {
			
			@Override
			public Boolean run() {
				ILayerCell bodyCell = getBodyLayer().getCellByPosition(col, row);
				if (bodyCell != null) {
					LabelStack labels = getBodyLayer().getCellByPosition(col, row).getConfigLabels();
				
					if (labels.getLabels().contains(label)) {
						return true;
					}
				}
				
				return false;
			}
		});
	}

	/**
	 * Triggers a mouse hover event on the given cell in the body region of the table,
	 * then waits for the specified delay (in ms).
	 * @param row row of the body cell
	 * @param col column of the body cell
	 * @param delay delay to wait in milliseconds
	 * @return true if hovering and waiting for the delay is successful, false otherwise
	 */
	public boolean hoverBodyCell(final int row, final int col, int delay) {
		// hover mouse to position of cell
		final Point p = getBodyCellCoordinates(row, col);
		notify(SWT.MouseHover, createMouseEvent(p.x, p.y, 0, 0, 0));
		// Wait for delay time
		try {
			Thread.sleep(delay);
		} catch (InterruptedException e) {
			return false;
		}
		return true;
	}

	/**
	 * Returns the text of a tooltip as created by the {@link DefaultToolTip} implementation.
	 * Assumes that the tooltip consists of a CLabel element within a separate shell that
	 * has the current RCP application shell as parent and defines certain styles.
	 * @param parentBot the {@link SWTBot} for the general application.
	 * @return
	 */
	public String getDefaultTooltipText(final SWTBot parentBot) {
		return syncExec(new StringResult() {
			@Override
			public String run() {
				// Get the general application shell
				SWTBotShell parentShell = parentBot.activeShell();
				SWTBotShell tooltipShell = null;
				// Search for the tooltip shell
				for (Shell shell : parentBot.shells("", parentShell.widget)) { //$NON-NLS-1$
					// Tooltip shell has the following styles
					if ((shell.getStyle() | SWT.ON_TOP | SWT.TOOL | SWT.NO_FOCUS) == shell.getStyle()) {
						tooltipShell = new SWTBotShell(shell);
					}
				}
				// Default tooltip is built of CLabel in a child shell of the application
				if (tooltipShell != null) {
					return tooltipShell.bot().clabel().getText();
				}
				return null;	
			}
		});
	}
	 
	/**
	 * Returns true if the given cell in the row header region has the expected label set,
	 * false otherwise.
	 */
	public boolean hasRowHeaderLabel(final int row, final String label) {
		return syncExec(new BoolResult() {
			
			@Override
			public Boolean run() {
				ILayer rowHeaderLayer = getRowHeaderLayer(); 
				if (rowHeaderLayer.getConfigLabelsByPosition(0, row).hasLabel(label)) {
					return true;
				}
				return false;
			}
		});
	}
	
	/**
	 * Sets the value to the specified cell directly on the data layer of the body region
	 * of the table by executing the update data command.
	 */
	public void setBodyValue(final int row, final int col, final Object value) {
		syncExec(new VoidResult() {
			@Override
			public void run() {
				ILayer bodyLayer = getBodyLayer();
				bodyLayer.doCommand(new UpdateDataCommand(bodyLayer, col, row, value));				
			}
		});
	}
	
	/**
	 * Short-cut for left click on body cell.
	 */
	public void clickBodyCell(final int row, final int col) {
		clickBodyCell(row, col, false);
	}
	
	/**
	 * Performs a click on the given cell in the body region of the table.
	 */
	public void clickBodyCell(final int row, final int col, final boolean rightClick) {
		// Find out coordinates for click
		Point p = getBodyCellCoordinates(row, col);
		
		if (rightClick) {
			rightClick(p.x, p.y, false);
		}
		else {
			clickXY(p.x, p.y);
		}
	}
	
	private Point getBodyCellCoordinates(final int row, final int col) {
		return syncExec(new Result<Point>() {

			@Override
			public Point run() {
				ILayer bodyLayer = getBodyLayer();
				Rectangle rec = bodyLayer.getBoundsByPosition(col, row);
				// Determine size of row and column header layers as an offset for the position
				ILayer rowHeaderLayer = getRowHeaderLayer();
				int startx = rowHeaderLayer!=null?rowHeaderLayer.getWidth():0;
				
				ILayer colHeaderLayer = getColumnHeaderLayer();
				int starty = colHeaderLayer != null?colHeaderLayer.getHeight():0;
				
				int x = startx + rec.x + rec.width/2;
				int y = starty + rec.y + rec.height/2;
				Point rp = new Point(x, y);
				return rp;
			}
		});
	}
	
	/**
	 * Performs a click on the header of the given row.
	 * @param row row number (0-based)
	 * @param rightClick true if a right click should be performed, false if a left click should be performed
	 */
	public boolean clickRowHeader(final int row, final boolean rightClick) {

		// Find out coordinates for click
		Point p = syncExec(new Result<Point>() {

			@Override
			public Point run() {
				ILayer headerLayer = getRowHeaderLayer();
				if (headerLayer == null) {
					return null;
				}
				Rectangle rec = headerLayer.getBoundsByPosition(0, row);
				if (rec == null) {
					return null;
				}
				// Determine size of column header layer as an offset for the position
				ILayer colHeaderLayer = getColumnHeaderLayer();
				int starty = colHeaderLayer != null?colHeaderLayer.getHeight():0;

				int x = rec.x + rec.width/2;
				int y = starty + rec.y + rec.height/2;
				Point rp = new Point(x, y);
				return rp;
			}
		});
		
		if (p == null) {
			return false;
		}
		
		if (rightClick) {
			rightClickXY(p.x, p.y);
		}
		else {
			clickXY(p.x, p.y);
		}
		return true;
	}
	
	/**
	 * Performs a click on the header of the given column.
	 * @param col column number (0-based)
	 * @param rightClick true if a right click should be performed, false if a left click should be performed
	 */
	public boolean clickColumnHeader(final int col, final boolean rightClick) {

		// Find out coordinates for click
		Point p = syncExec(new Result<Point>() {

			@Override
			public Point run() {
				ILayer colHeaderLayer = getColumnHeaderLayer();
				if (colHeaderLayer == null) {
					return null;
				}
				Rectangle rec = colHeaderLayer.getBoundsByPosition(col, 0);
				if (rec == null) {
					return null;
				}
				// Determine size of column header layer as an offset for the position
				ILayer rowHeaderLayer = getRowHeaderLayer();
				int startx = rowHeaderLayer != null?rowHeaderLayer.getHeight():0;

				int x = startx + rec.x + rec.width/2;
				int y = rec.y + rec.height/2;
				Point rp = new Point(x, y);
				return rp;
			}
		});
		
		if (p == null) {
			return false;
		}
		
		if (rightClick) {
			rightClickXY(p.x, p.y);
		}
		else {
			clickXY(p.x, p.y);
		}
		return true;
	}

	
	/**
	 * Return the context menu with the specified text
	 */
	public SWTBotMenu contextMenu(String text) {
		return contextMenu(widget, text);
	}
	
	/*
	 * Corrected implementation of the clickXY method in the super class that performs a proper right click.
	 */
	private void rightClickXY(int x, int y) {
		notify(SWT.MouseEnter);
		notify(SWT.MouseMove);
		notify(SWT.Activate);
		notify(SWT.FocusIn);
		notify(SWT.MouseDown, createMouseEvent(x, y, 3, SWT.NONE, 1));
		notify(SWT.MouseUp, createMouseEvent(x, y, 3, SWT.BUTTON1, 1));
		notify(SWT.Selection, createSelectionEvent(SWT.BUTTON1));
		notify(SWT.MouseHover);
		notify(SWT.MouseMove);
		notify(SWT.MouseExit);
		notify(SWT.Deactivate);
		notify(SWT.FocusOut);
	}
	
	/**
	 * Selects the entry with the index in the combo box editor of the given cell of the body region.
	 * @param row
	 * @param col
	 * @param index
	 */
	public boolean selectBodyComboBoxValue(final int row, final int col, final int index) {
		return syncExec(new BoolResult() {

			@Override
			public Boolean run() {
				ILayerCell cell = getBodyLayer().getCellByPosition(col, row);
				if (cell == null) {
					return false;
				}
				LabelStack labels = cell.getConfigLabels();
				
				// get the cell editor from the label
				ICellEditor editor = widget.getConfigRegistry().getConfigAttribute(EditConfigAttributes.CELL_EDITOR,
						DisplayMode.EDIT, labels.getLabels());
				// we expect a combobox celleditor here
				if (editor instanceof ComboBoxCellEditor) {
					ComboBoxCellEditor comboEditor = ((ComboBoxCellEditor)editor);
					comboEditor.activateCell(widget, null, EditModeEnum.INLINE, 
							new InlineEditHandler(getBodyLayer(), col, row), cell, widget.getConfigRegistry());
					comboEditor.select(index);
					comboEditor.commit(MoveDirectionEnum.RIGHT);
					return true;
				}
				return false;
			}
		});
	}
	
	/**
	 * Returns a condition that checks whether the number of rows in the body region
	 * corresponds to the expected count value.
	 * @param expectedCount the expected number of rows
	 * @return condition object to be used by test case
	 */
	public ICondition getBodyRowCountCondition(int expectedCount) {
		return new BodyRowCountCondition(this, expectedCount);
	}

	/**
	 * Returns a condition that checks whether the value of a particular cell
	 * in the body region corresponds to the expected String value.
	 * @param row the row number of the cell (0-based)
	 * @param col the column number of the cell (0-based)
	 * @param expectedValue the expected value
	 * @return condition object
	 */
	public ICondition getBodyStringValueCondition(int row, int col, String expectedValue) {
		return new BodyStringValueCondition(this, row, col, expectedValue);
	}

	/**
	 * Returns a condition that checks whether a particular cell in the body region
	 * has an expected label set
	 * @param row the row number of the cell (0-based)
	 * @param col the column number of the cell (0-based)
	 * @param expectedLabel the expected label
	 * @return condition object
	 */
	public ICondition getBodyLabelCondition(int row, int col, String expectedLabel) {
		return new BodyLabelCondition(this, row, col, expectedLabel);
	}
	
	/**
	 * Returns a condition the checks whether a particular rowheader has an expected
	 * label set.
	 * @param row the row number of the cell (0-based)
	 * @param expectedLabel the expected label	 * 
	 * @return condition object
	 */
	public ICondition getRowHeaderLabelCondition(int row, String expectedLabel) {
		return new RowHeaderLabelCondition(this, row, expectedLabel);
	}

	
	/**
	 * Returns a condition that checks whether the value of a particular cell
	 * in the body region corresponds to the expected float value.
	 * @param row the row number of the cell (0-based)
	 * @param col the column number of the cell (0-based)
	 * @param expectedValue the expected value
	 * @param epsilon maximum difference between actual and expected value (to consider rounding differences)
	 * @return
	 */
	public ICondition getBodyFloatValueCondition(int row, int col, float expectedValue, float epsilon) {
		return new BodyFloatValueCondition(this, row, col, expectedValue, epsilon);
	}

	
	/**
	 * Condition that evaluates to true if the body layer of the associated table from the bot
	 * holds the expected number of rows.
	 */
	private class BodyRowCountCondition extends DefaultCondition {
		
		private SWTBotNatTable bot;
		private int expectedCount;
		
		public BodyRowCountCondition(SWTBotNatTable bot, int expectedCount) {
			this.bot = bot;
			this.expectedCount = expectedCount;
		}

		@SuppressWarnings("nls")
		@Override
		public String getFailureMessage() {
			return "Count not verified: expected " + expectedCount + ", actual " + bot.getBodyRowCount();
		}

		@Override
		public boolean test() throws Exception {
			return bot.getBodyRowCount() == expectedCount;
		}
	}
	
	/**
	 * Condition that evaluates to true if the cell at the specified row and column position of the
	 * body layer of the associated table from the bot has the expected label assigned.
	 */
	private class BodyLabelCondition extends DefaultCondition {
		
		private SWTBotNatTable bot;
		private int row, col;
		private String expectedLabel;
		
		public BodyLabelCondition(SWTBotNatTable bot, int row, int col, String expectedLabel) {
			this.bot = bot;
			this.row = row;
			this.col = col;
			this.expectedLabel = expectedLabel;
		}

		@Override
		public String getFailureMessage() {
			return "Label not verified"; //$NON-NLS-1$
		}

		@Override
		public boolean test() throws Exception {
			return bot.hasBodyCellLabel(row, col, expectedLabel);
		}
	}
	
	/**
	 * Condition that evaluates to true if the row header of the specified row
	 * of the associated table from the bot has the expected label assigned.
	 */
	private class RowHeaderLabelCondition extends DefaultCondition {
		
		private SWTBotNatTable bot;
		private int row;
		private String expectedLabel;
		
		public RowHeaderLabelCondition(SWTBotNatTable bot, int row, String expectedLabel) {
			this.bot = bot;
			this.row = row;
			this.expectedLabel = expectedLabel;
		}

		@Override
		public String getFailureMessage() {
			return "Row Header Label not verified"; //$NON-NLS-1$
		}

		@Override
		public boolean test() throws Exception {
			return bot.hasRowHeaderLabel(row, expectedLabel);
		}
	}

	
	/**
	 * Condition that evaluates to true as soon as the String value of the cell
	 * with the given row and column in the table bot passed in bot equals the expected value.
	 */
	private class BodyStringValueCondition extends DefaultCondition {
		
		private SWTBotNatTable bot;
		private String expectedValue;
		private int row, col;
		
		public BodyStringValueCondition(SWTBotNatTable bot, int row, int col, String expectedValue) {
			this.bot = bot;
			this.row = row;
			this.col = col;
			this.expectedValue = expectedValue;
		}

		@SuppressWarnings("nls")
		@Override
		public String getFailureMessage() {
			Object value = bot.getBodyValue(row, col);
			return "Value not verified: expected " + expectedValue + ", actual " + value!=null?value.toString():"null";
		}

		@Override
		public boolean test() throws Exception {
			Object value = bot.getBodyValue(row, col);
			return (value != null && value.toString().equals(expectedValue));
		}
	}
	
	/**
	 * Condition that evaluates to true as soon as the String value of the cell
	 * with the given row and column in the table bot passed equals the expected value.
	 */
	private class BodyFloatValueCondition extends DefaultCondition {
		
		private SWTBotNatTable bot;
		private float expectedValue;
		private float epsilon;
		private int row, col;
		
		public BodyFloatValueCondition(SWTBotNatTable bot, int row, int col, float expectedValue, float epsilon) {
			this.bot = bot;
			this.row = row;
			this.col = col;
			this.expectedValue = expectedValue;
			this.epsilon = epsilon;
		}

		@SuppressWarnings("nls")
		@Override
		public String getFailureMessage() {
			Object value = bot.getBodyValue(row, col);
			String actualValue = "null";
			if (actualValue != null) {
				if (value instanceof Float) {
					actualValue = String.valueOf(value);
				}
				else {
					actualValue = "Not of type Float";
				}
			}
			
			return "Value not verified: expected " + expectedValue + String.valueOf(expectedValue) + ", actual " + actualValue;
		}

		@Override
		public boolean test() throws Exception {
			Object value = bot.getBodyValue(row, col);
			return value instanceof Float && Math.abs((Float)value - expectedValue) < epsilon;
		}
	}
}
