/*
Copyright (c) 2013, BeOne Stuttgart GmbH
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
- Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer
 in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.beone.sample.test;

import static org.eclipse.swtbot.swt.finder.matchers.WidgetMatcherFactory.widgetOfType;

import org.eclipse.nebula.widgets.nattable.NatTable;
import org.eclipse.swtbot.eclipse.finder.SWTBotEclipseTestCase;
import org.eclipse.swtbot.swt.finder.widgets.SWTBotMenu;
import org.junit.Test;

import com.beone.sample.View;

/**
 * Sample Test class to show the capabilities for the {@link SWTBotNatTable} extension.
 */
public class SampleTest extends SWTBotEclipseTestCase {
	
	@Test
	public void testSampleTable() throws Exception {
		// Load SWTBotNatTable instance
		SWTBotNatTable botNatTable = new SWTBotNatTable(
				bot.widget(widgetOfType(NatTable.class)));
		// Verify number of rows and columns
		assertEquals(2, botNatTable.getBodyRowCount());
		assertEquals(5, botNatTable.getBodyColumnCount());
		
		// Add new element, wait for number of rows to increase
		bot.button("Add new table element").click();
		bot.waitUntil(botNatTable.getBodyRowCountCondition(3));
		
		// Set values of new entry
		botNatTable.setBodyValue(2, 1, "Vogel");
		botNatTable.setBodyValue(2, 2, "Lars");
		
		// Verify values
		assertEquals("Lars", botNatTable.getBodyValue(2, 2));
		assertEquals("beginner", botNatTable.getBodyValue(2, 3));
		assertEquals(false, botNatTable.getBodyValue(2, 4));
		// no label set for level column
		assertFalse(botNatTable.hasBodyCellLabel(2, 3, View.EXPERT_LABEL));
		// verify tooltip of 3rd column in second row:
		// hover for 100ms more than the default delay of 0,5 sec
		botNatTable.hoverBodyCell(1, 2, 600);
		assertEquals("Dirk", botNatTable.getDefaultTooltipText(bot));
		
		// Select 3rd value (expert) in combo box
		botNatTable.selectBodyComboBoxValue(2, 3, 2);
		
		// assert new value and that expert label is set
		assertEquals("expert", botNatTable.getBodyValue(2, 3));
		// now label is set for level column
		assertTrue("Expert Label not set", 
				botNatTable.hasBodyCellLabel(2, 3, View.EXPERT_LABEL));
		
		// Click (and thus check) the checkbox in the 5th column
		botNatTable.clickBodyCell(2, 4);
		assertEquals("Value not checked", true, botNatTable.getBodyValue(2, 4));
		
		// Perform right click on first line row header (necessary to show context menu)
		botNatTable.clickRowHeader(0, true);
		// Delete second entry via context menu
		SWTBotMenu menuItem = botNatTable.contextMenu("Delete entry");
		assertNotNull(menuItem);
		menuItem.click();
				
		// Verify number of rows has decreased
		bot.waitUntil(botNatTable.getBodyRowCountCondition(2));
		// Verify last names of remaining elements
		assertEquals("Fauth", botNatTable.getBodyValue(0, 1));
		assertEquals("Vogel", botNatTable.getBodyValue(1, 1));
	}
}
